// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{ops::BundleLiteral, Expr, ToExpr},
    intern::{Intern, Interned},
    module::{ModuleBuilder, NormalModule},
    source_location::SourceLocation,
    ty::{
        impl_match_variant_as_self, CanonicalType, MatchVariantWithoutScope, StaticType, Type,
        TypeProperties, TypeWithDeref,
    },
};
use hashbrown::HashMap;

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq)]
pub struct BundleField {
    pub name: Interned<str>,
    pub flipped: bool,
    pub ty: CanonicalType,
}

#[derive(Clone, Eq)]
struct BundleImpl {
    fields: Interned<[BundleField]>,
    name_indexes: HashMap<Interned<str>, usize>,
    field_offsets: Interned<[usize]>,
    type_properties: TypeProperties,
}

impl std::hash::Hash for BundleImpl {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.fields.hash(state);
    }
}

impl PartialEq for BundleImpl {
    fn eq(&self, other: &Self) -> bool {
        self.fields == other.fields
    }
}

impl std::fmt::Debug for BundleImpl {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Bundle")
            .field("fields", &self.fields)
            .field("name_indexes", &self.name_indexes)
            .field("field_offsets", &self.field_offsets)
            .field("type_properties", &self.type_properties)
            .finish()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct Bundle(Interned<BundleImpl>);

impl std::fmt::Debug for Bundle {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl Bundle {
    #[track_caller]
    pub fn new(fields: Interned<[BundleField]>) -> Self {
        let mut name_indexes = HashMap::with_capacity(fields.len());
        let mut field_offsets = Vec::with_capacity(fields.len());
        let mut type_properties = TypeProperties {
            is_passive: true,
            is_storable: true,
            is_castable_from_bits: true,
            bit_width: 0,
        };
        for (index, &BundleField { name, flipped, ty }) in fields.iter().enumerate() {
            if let Some(old_index) = name_indexes.insert(name, index) {
                panic!("duplicate field name {name:?}: at both index {old_index} and {index}");
            }
            field_offsets.push(type_properties.bit_width);
            let bit_width = type_properties
                .bit_width
                .checked_add(ty.bit_width())
                .unwrap_or_else(|| panic!("bundle is too big: bit-width overflowed"));
            type_properties = if flipped {
                TypeProperties {
                    is_passive: false,
                    is_storable: false,
                    is_castable_from_bits: false,
                    bit_width,
                }
            } else {
                let field_props = ty.type_properties();
                TypeProperties {
                    is_passive: type_properties.is_passive & field_props.is_passive,
                    is_storable: type_properties.is_storable & field_props.is_storable,
                    is_castable_from_bits: type_properties.is_castable_from_bits
                        & field_props.is_castable_from_bits,
                    bit_width,
                }
            };
        }
        Self(Intern::intern_sized(BundleImpl {
            fields,
            name_indexes,
            field_offsets: Intern::intern_owned(field_offsets),
            type_properties,
        }))
    }
    pub fn name_indexes(&self) -> &HashMap<Interned<str>, usize> {
        &self.0.name_indexes
    }
    pub fn field_by_name(&self, name: Interned<str>) -> Option<BundleField> {
        Some(self.0.fields[*self.0.name_indexes.get(&name)?])
    }
    pub fn field_offsets(self) -> Interned<[usize]> {
        self.0.field_offsets
    }
    pub fn type_properties(self) -> TypeProperties {
        self.0.type_properties
    }
}

impl Type for Bundle {
    type MaskType = Bundle;
    impl_match_variant_as_self!();
    fn mask_type(&self) -> Self::MaskType {
        Self::new(Interned::from_iter(self.0.fields.into_iter().map(
            |BundleField { name, flipped, ty }| BundleField {
                name,
                flipped,
                ty: ty.mask_type(),
            },
        )))
    }
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Bundle(*self)
    }
    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Bundle(bundle) = canonical_type else {
            panic!("expected bundle");
        };
        bundle
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

pub trait BundleType: Type {
    fn fields(&self) -> Interned<[BundleField]>;
}

impl BundleType for Bundle {
    fn fields(&self) -> Interned<[BundleField]> {
        self.0.fields
    }
}

macro_rules! impl_tuples {
    ([$({#[num = $num:literal] $var:ident: $T:ident})*] []) => {
        impl<$($T: Type,)*> Type for ($($T,)*) {
            type MaskType = ($($T::MaskType,)*);
            type MatchVariant = ($(Expr<$T>,)*);
            type MatchActiveScope = ();
            type MatchVariantAndInactiveScope = MatchVariantWithoutScope<Self::MatchVariant>;
            type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;
            fn match_variants(
                this: Expr<Self>,
                module_builder: &mut ModuleBuilder<NormalModule>,
                source_location: SourceLocation,
            ) -> Self::MatchVariantsIter {
                let _ = this;
                let _ = module_builder;
                let _ = source_location;
                std::iter::once(MatchVariantWithoutScope(($(Expr::field(this, stringify!($num)),)*)))
            }
            fn mask_type(&self) -> Self::MaskType {
                let ($($var,)*) = self;
                ($($var.mask_type(),)*)
            }
            fn canonical(&self) -> CanonicalType {
                Bundle::new(self.fields()).canonical()
            }
            #[track_caller]
            fn from_canonical(canonical_type: CanonicalType) -> Self {
                let CanonicalType::Bundle(bundle) = canonical_type else {
                    panic!("expected bundle");
                };
                let [$($var,)*] = *bundle.fields() else {
                    panic!("bundle has wrong number of fields");
                };
                $(let BundleField { name, flipped, ty } = $var;
                assert_eq!(&*name, stringify!($num));
                assert_eq!(flipped, false);
                let $var = $T::from_canonical(ty);)*
                ($($var,)*)
            }
            fn source_location() -> SourceLocation {
                SourceLocation::builtin()
            }
        }
        impl<$($T: Type,)*> BundleType for ($($T,)*) {
            fn fields(&self) -> Interned<[BundleField]> {
                let ($($var,)*) = self;
                [$(BundleField { name: stringify!($num).intern(), flipped: false, ty: $var.canonical() }),*][..].intern()
            }
        }
        impl<$($T: Type,)*> TypeWithDeref for ($($T,)*) {
            fn expr_deref(this: &Expr<Self>) -> &Self::MatchVariant {
                let _ = this;
                Interned::<_>::into_inner(($(Expr::field(*this, stringify!($num)),)*).intern_sized())
            }
        }
        impl<$($T: StaticType,)*> StaticType for ($($T,)*) {
            fn static_type() -> Self {
                $(let $var = $T::static_type();)*
                ($($var,)*)
            }
        }
        impl<$($T: ToExpr,)*> ToExpr for ($($T,)*) {
            type Type = ($($T::Type,)*);

            fn to_expr(&self) -> Expr<Self::Type> {
                let ($($var,)*) = self;
                $(let $var = $var.to_expr();)*
                let ty = ($(Expr::ty($var),)*);
                let field_values = [$(Expr::canonical($var)),*];
                BundleLiteral::new(ty, &field_values).to_expr()
            }
        }
    };
    ([$($lhs:tt)*] [$rhs_first:tt $($rhs:tt)*]) => {
        impl_tuples!([$($lhs)*] []);
        impl_tuples!([$($lhs)* $rhs_first] [$($rhs)*]);
    };
}

impl_tuples! {
    [] [
        {#[num = 0] a: A}
        {#[num = 1] b: B}
        {#[num = 2] c: C}
        {#[num = 3] d: D}
        {#[num = 4] e: E}
        {#[num = 5] f: F}
        {#[num = 6] g: G}
        {#[num = 7] h: H}
        {#[num = 8] i: I}
        {#[num = 9] j: J}
        {#[num = 10] k: K}
        {#[num = 11] l: L}
    ]
}
